"""CLI for non-diffeomorphic volume and non-diffeomorphic area
Author: Yihao Liu <yliu236@jhu.edu>
Created on: April 2023
"""
import argparse
import nibabel as nib
import numpy as np
import torch
import os
import logging

from digital_diffeomorphism.core import ndv, nda

def non_diffeomorphic_volume():
    parser = argparse.ArgumentParser(description='Calculate non-diffeomorphic volume.')
    parser.add_argument('trans_file', type=str, help='Path to the transformation Nifti file')
    parser.add_argument('--disp', action='store_true', help='Flag for displacement field input')
    parser.add_argument('--mask_file', type=str, help='Path to the mask Nifti file')
    parser.add_argument('--save_results', action='store_true', help='Save the output nifti in the transformation folder')
    parser.add_argument('--gpu', action='store_true', help='Flag for using GPU')
    parser.add_argument('--verbose', action='store_true', help='Increase output verbosity')
    args = parser.parse_args()

    # Set logging level based on verbose argument
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    logging.debug('Start data loading and pre-processing...')
    # load transformation data
    trans_obj = nib.load(args.trans_file)
    trans, affine = trans_obj.get_fdata(), trans_obj.affine

    if args.mask_file:
        mask = nib.load(args.mask_file).get_fdata()
    else:
        mask = None

    logging.debug('Calculating NDV...')
    nd_volume, nd_volume_pct, nd_volume_map, nd_voxels, nd_voxels_pct, nd_voxels_map = ndv(trans, mask, args.disp, args.gpu)

    print(f'Non-diffeomorphic Volume: {nd_volume:.2f}({nd_volume_pct:.2f}%)')
    print(f'Non-diffeomorphic Voxels: {nd_voxels:.2f}({nd_voxels_pct:.2f}%)')

    if args.save_results:
        save_filename = args.trans_file.replace('.nii.gz', '_non_diff_volume.nii.gz')
        nib.save(nib.Nifti1Image(nd_volume_map, affine), save_filename)
        logging.info(f'Saved non-diffeomorphic volume map to {save_filename}')

        save_filename = args.trans_file.replace('.nii.gz', '_non_diff_voxels.nii.gz')
        nib.save(nib.Nifti1Image(nd_voxels_map, affine), save_filename)
        logging.info(f'Saved non-diffeomorphic voxel map to {save_filename}')

def non_diffeomorphic_area():
    parser = argparse.ArgumentParser(description='Calculate non-diffeomorphic area')
    parser.add_argument('trans_file', type=str, help='Path to the transformation Nifti file')
    parser.add_argument('--mask_file', type=str, help='Path to the mask Nifti file')
    parser.add_argument('--disp', action='store_true', help='Flag for displacement field input')
    parser.add_argument('--save_results', action='store_true', help='Save the output in the transformation folder')
    parser.add_argument('--gpu', action='store_true', help='Flag for using GPU')
    parser.add_argument('--verbose', action='store_true', help='Increase output verbosity')
    args = parser.parse_args()

    # Set logging level based on verbose argument
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    logging.debug('Start data loading and pre-processing...')
    # load transformation data
    trans_obj = nib.load(args.trans_file)
    trans, affine = trans_obj.get_fdata(), trans_obj.affine

    if args.mask_file:
        mask = nib.load(args.mask_file).get_fdata()
    else:
        mask = None

    logging.debug('Calculating NDA...')
    nd_area, nd_area_pct, nd_area_map, nd_pixels, nd_pixels_pct, nd_pixels_map = nda(trans, mask, args.disp, args.gpu)

    print(f'Non-diffeomorphic Area: {nd_area:.2f}({nd_area_pct:.2f}%)')
    print(f'Non-diffeomorphic Pixels: {nd_pixels:.2f}({nd_pixels_pct:.2f}%)')

    if args.save_results:
        save_filename = args.trans_file.replace('.nii.gz', '_non_diff_area.nii.gz')
        nib.save(nib.Nifti1Image(nd_area_map, affine), save_filename)
        logging.info(f'Saved non-diffeomorphic area map to {save_filename}')

        save_filename = args.trans_file.replace('.nii.gz', '_non_diff_pixels.nii.gz')
        nib.save(nib.Nifti1Image(nd_pixels_map, affine), save_filename)
        logging.info(f'Saved non-diffeomorphic pixels map to {save_filename}')




"""Core functions for non-diffeomorphic volume and non-diffeomorphic area
Author: Yihao Liu <yihao.liu@vanderbilt.edu>
Created on: March 2025
"""
import argparse
import nibabel as nib
import numpy as np
import torch
import os
import logging

from digital_diffeomorphism.jacobian_calculator_torch import (
                    calc_jacdets_3d,
                    calc_jacdets_2d
)

from digital_diffeomorphism.non_diffeomorphic_measures import (
                    calc_non_diffeomorphic_volume,
                    calc_non_diffeomorphic_voxels,
                    calc_non_diffeomorphic_area,
                    calc_non_diffeomorphic_pixels
)

from digital_diffeomorphism.utils import identity_grid_like, spatial_padding


def ndv(trans, mask=None, is_disp=False, use_gpu=False):
    # convert displacement field to deformation field if necessary
    if is_disp:
        trans += identity_grid_like(trans)

    if use_gpu:
        os.environ["CUDA_VISIBLE_DEVICES"] = '0'
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    else:
        device = torch.device("cpu")

    jacdets = calc_jacdets_3d(trans, device)

    if mask is not None:
        mask = (mask > 0).astype('float32')
    else:
        mask = np.ones_like(trans[..., 0])

    total_voxels = np.sum(mask)

    non_diff_volume_map = calc_non_diffeomorphic_volume(jacdets, mask)
    non_diff_volume = np.sum(non_diff_volume_map)
    non_diff_volume_pct = non_diff_volume / total_voxels * 100

    non_diff_voxels_map = calc_non_diffeomorphic_voxels(jacdets, mask)
    non_diff_voxels = np.sum(non_diff_voxels_map)
    non_diff_voxels_pct = non_diff_voxels / total_voxels * 100

    return non_diff_volume, non_diff_volume_pct, non_diff_volume_map, non_diff_voxels, non_diff_voxels_pct, non_diff_voxels_map

def nda(trans, mask=None, is_disp=False, use_gpu=False):
    # convert displacement field to deformation field if necessary
    if is_disp:
        trans += identity_grid_like(trans)

    if use_gpu:
        os.environ["CUDA_VISIBLE_DEVICES"] = '0'
        device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    else:
        device = torch.device("cpu")

    jacdets = calc_jacdets_2d(trans, device)

    if mask is not None:
        mask = (mask > 0).astype('float32')
    else:
        mask = np.ones_like(trans[..., 0])

    total_pixels = np.sum(mask)

    non_diff_area_map = calc_non_diffeomorphic_area(jacdets, mask)
    non_diff_area = np.sum(non_diff_area_map)
    non_diff_area_pct = non_diff_area / total_pixels * 100

    non_diff_pixels_map = calc_non_diffeomorphic_pixels(jacdets, mask)
    non_diff_pixels = np.sum(non_diff_pixels_map)
    non_diff_pixels_pct = non_diff_pixels / total_pixels * 100

    return non_diff_area, non_diff_area_pct, non_diff_area_map, non_diff_pixels, non_diff_pixels_pct, non_diff_pixels_map



"""Top-level package for digital-diffeomorphism."""
from digital_diffeomorphism.core import ndv, nda

__title__ = "digital-diffeomorphism"
__description__ = "Digital diffeomorphism: Non-diffeomorphic volume and non-diffeomorphic area computation"
__url__ = "https://github.com/yihao6/digital_diffeomorphism"
__author__ = """Yihao Liu"""
__email__ = "yliu236@jhu.edu"
__version__ = "1.1.1"
__license__ = "GPLv3"
__copyright__ = "Copyright 2023 Yihao Liu"



"""Calculate Jacobian determinants
Author: Yihao Liu <yliu236@jhu.edu>
Created on: November 2023
"""
import numpy as np
import torch
import torch.nn.functional as nnf
from torch.linalg import det

from digital_diffeomorphism.utils import spatial_padding

def calc_jacdets_3d(trans, device):

    if trans.ndim != 4 or trans.shape[3] != 3:
        raise ValueError("Input transformation must have shape (H, W, D, 3).")

    trans = spatial_padding(trans)

    kernels = {}
    kwargs = {'dtype': torch.float32, 'device': device}
    kernels['D0x'] = torch.tensor([-0.5,   0, 0.5], **kwargs).view(1, 1, 3, 1, 1)
    kernels['D+x'] = torch.tensor([   0,  -1,   1], **kwargs).view(1, 1, 3, 1, 1)
    kernels['D-x'] = torch.tensor([  -1,   1,   0], **kwargs).view(1, 1, 3, 1, 1)

    kernels['D0y'] = torch.tensor([-0.5,   0, 0.5], **kwargs).view(1, 1, 1, 3, 1)
    kernels['D+y'] = torch.tensor([   0,  -1,   1], **kwargs).view(1, 1, 1, 3, 1)
    kernels['D-y'] = torch.tensor([  -1,   1,   0], **kwargs).view(1, 1, 1, 3, 1)

    kernels['D0z'] = torch.tensor([-0.5,   0, 0.5], **kwargs).view(1, 1, 1, 1, 3)
    kernels['D+z'] = torch.tensor([   0,  -1,   1], **kwargs).view(1, 1, 1, 1, 3)
    kernels['D-z'] = torch.tensor([  -1,   1,   0], **kwargs).view(1, 1, 1, 1, 3)

    # J1*
    kernels['1*xy'] = torch.tensor(
                            [[1, 0, 0], [0, -1, 0], [0, 0, 0]],
                            **kwargs,
    ).reshape(1, 1, 3, 3, 1)
    kernels['1*xz'] = torch.tensor(
                            [[1, 0, 0], [0, -1, 0], [0, 0, 0]],
                            **kwargs,
    ).reshape(1, 1, 3, 1, 3)
    kernels['1*yz'] = torch.tensor(
                            [[1, 0, 0], [0, -1, 0], [0, 0, 0]],
                            **kwargs,
    ).reshape(1, 1, 1, 3, 3)

    # J2*
    kernels['2*xy'] = torch.tensor(
                            [[0, 0, 0], [0, -1, 0], [0, 0, 1]],
                            **kwargs,
    ).reshape(1, 1, 3, 3, 1)
    kernels['2*xz'] = torch.tensor(
                            [[0, 0, 0], [0, -1, 0], [0, 0, 1]],
                            **kwargs,
    ).reshape(1, 1, 3, 1, 3)
    kernels['2*yz'] = torch.tensor(
                            [[0, 0, 0], [0, -1, 0], [0, 0, 1]],
                            **kwargs,
    ).reshape(1, 1, 1, 3, 3)

    for kernel_name, kernel in kernels.items():
        kernels[kernel_name] = kernel.type(torch.float32)

    # combine kernels with the same sizes
    weights = {
        'x': torch.cat([kernels[key] for key in ['D0x', 'D+x', 'D-x']] * 3, dim=0),
        'y': torch.cat([kernels[key] for key in ['D0y', 'D+y', 'D-y']] * 3, dim=0),
        'z': torch.cat([kernels[key] for key in ['D0z', 'D+z', 'D-z']] * 3, dim=0),

        '*xy': torch.cat([kernels[key] for key in ['1*xy', '2*xy']] * 3, dim=0),
        '*xz': torch.cat([kernels[key] for key in ['1*xz', '2*xz']] * 3, dim=0),
        '*yz': torch.cat([kernels[key] for key in ['1*yz', '2*yz']] * 3, dim=0),
    }

    trans = torch.from_numpy(trans).permute(3, 0, 1, 2)
    trans = (trans[None,...]).type(torch.float32).to(device)

    partials = {
        'x': nnf.conv3d(trans, weights['x'], groups=3)[:, :, :, 1:-1, 1:-1],
        'y': nnf.conv3d(trans, weights['y'], groups=3)[:, :, 1:-1, :, 1:-1],
        'z': nnf.conv3d(trans, weights['z'], groups=3)[:, :, 1:-1, 1:-1, :],

        '*xy': nnf.conv3d(trans, weights['*xy'], groups=3)[:, :, :, :, 1:-1],
        '*xz': nnf.conv3d(trans, weights['*xz'], groups=3)[:, :, :, 1:-1, :],
        '*yz': nnf.conv3d(trans, weights['*yz'], groups=3)[:, :, 1:-1, :, :],
    }

    jacobians = {
        '000': torch.stack((
                            partials['x'][:, ::3, ...],
                            partials['y'][:, ::3, ...],
                            partials['z'][:, ::3, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
        '+++': torch.stack((
                            partials['x'][:, 1::3, ...],
                            partials['y'][:, 1::3, ...],
                            partials['z'][:, 1::3, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
        '++-': torch.stack((
                            partials['x'][:, 1::3, ...],
                            partials['y'][:, 1::3, ...],
                            partials['z'][:, 2::3, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
        '+-+': torch.stack((
                            partials['x'][:, 1::3, ...],
                            partials['y'][:, 2::3, ...],
                            partials['z'][:, 1::3, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
        '+--': torch.stack((
                            partials['x'][:, 1::3, ...],
                            partials['y'][:, 2::3, ...],
                            partials['z'][:, 2::3, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
        '-++': torch.stack((
                            partials['x'][:, 2::3, ...],
                            partials['y'][:, 1::3, ...],
                            partials['z'][:, 1::3, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
        '-+-': torch.stack((
                            partials['x'][:, 2::3, ...],
                            partials['y'][:, 1::3, ...],
                            partials['z'][:, 2::3, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
        '--+': torch.stack((
                            partials['x'][:, 2::3, ...],
                            partials['y'][:, 2::3, ...],
                            partials['z'][:, 1::3, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
        '---': torch.stack((
                            partials['x'][:, 2::3, ...],
                            partials['y'][:, 2::3, ...],
                            partials['z'][:, 2::3, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
        'j1*': torch.stack((
                            partials['*xy'][:, 0::2, ...],
                            partials['*xz'][:, 0::2, ...],
                            partials['*yz'][:, 0::2, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
        'j2*': torch.stack((
                            partials['*xy'][:, 1::2, ...],
                            partials['*yz'][:, 1::2, ...],
                            partials['*xz'][:, 1::2, ...],
                ), dim=-1).permute(0, 2, 3, 4, 1, 5),
    }

    jacdets = {key: det(jacobians[key]).cpu().numpy().squeeze() for key in jacobians}

    return jacdets

def calc_jacdets_2d(trans, device):

    if trans.ndim != 3 or trans.shape[2] != 2:
        raise ValueError("Input transformation must have shape (H, W, 2).")

    trans = spatial_padding(trans)

    kernels = {}
    kwargs = {'dtype': torch.float32, 'device': device}
    kernels['D0x'] = torch.tensor([-0.5,   0, 0.5], **kwargs).view(1, 1, 3, 1)
    kernels['D+x'] = torch.tensor([   0,  -1,   1], **kwargs).view(1, 1, 3, 1)
    kernels['D-x'] = torch.tensor([  -1,   1,   0], **kwargs).view(1, 1, 3, 1)

    kernels['D0y'] = torch.tensor([-0.5,   0, 0.5], **kwargs).view(1, 1, 1, 3)
    kernels['D+y'] = torch.tensor([   0,  -1,   1], **kwargs).view(1, 1, 1, 3)
    kernels['D-y'] = torch.tensor([  -1,   1,   0], **kwargs).view(1, 1, 1, 3)

    for kernel_name, kernel in kernels.items():
        kernels[kernel_name] = kernel.type(torch.float32)

    # combine kernels with the same sizes
    weights = {
        'x': torch.cat([kernels[key] for key in ['D0x', 'D+x', 'D-x']] * 2, dim=0),
        'y': torch.cat([kernels[key] for key in ['D0y', 'D+y', 'D-y']] * 2, dim=0),
    }

    trans = torch.from_numpy(trans).permute(2, 0, 1)
    trans = (trans[None,...]).type(torch.float32).to(device)

    partials = {
        'x': nnf.conv2d(trans, weights['x'], groups=2)[:, :, :, 1:-1],
        'y': nnf.conv2d(trans, weights['y'], groups=2)[:, :, 1:-1, :],
    }

    jacobians = {
        '00': torch.stack((
                            partials['x'][:, ::3, ...],
                            partials['y'][:, ::3, ...],
                ), dim=-1).permute(0, 2, 3, 1, 4),
        '++': torch.stack((
                            partials['x'][:, 1::3, ...],
                            partials['y'][:, 1::3, ...],
                ), dim=-1).permute(0, 2, 3, 1, 4),
        '+-': torch.stack((
                            partials['x'][:, 1::3, ...],
                            partials['y'][:, 2::3, ...],
                ), dim=-1).permute(0, 2, 3, 1, 4),
        '-+': torch.stack((
                            partials['x'][:, 2::3, ...],
                            partials['y'][:, 1::3, ...],
                ), dim=-1).permute(0, 2, 3, 1, 4),
        '--': torch.stack((
                            partials['x'][:, 2::3, ...],
                            partials['y'][:, 2::3, ...],
                ), dim=-1).permute(0, 2, 3, 1, 4),
    }

    jacdets = {key: det(jacobians[key]).cpu().numpy().squeeze() for key in jacobians}

    return jacdets




"""Calculate non-diffeomorphic measures
Author: Yihao Liu <yliu236@jhu.edu>
Created on: April 2023
"""
import numpy as np

def calc_non_diffeomorphic_volume(jacobian_determinants, mask):
    """
    Calculates the non-diffeomorphic volume using the given jacobian determinants and mask.

    Parameters:
    ----------
    jacobian_determinants : dict
        A dictionary of jacobian determinants for different directions.
    mask : np.ndarray
        A binary mask indicating the region of interest.

    Returns:
    -------
    non_diff_volume : np.ndarray
        A matrix of floating numbers that indicate the non-diffeomorphic volume.
    """
    non_diff_volume = np.zeros_like(jacobian_determinants['000'])
    for diff_direction in ['+++', '++-', '+-+', '+--', 'j1*', 'j2*', '-++', '-+-', '--+', '---']:
        non_diff_volume += -0.5 * np.minimum(jacobian_determinants[diff_direction], 0) * mask / 6

    return non_diff_volume

def calc_non_diffeomorphic_area(jacobian_determinants, mask):
    """
    Calculates the non-diffeomorphic area using the given jacobian determinants and mask.

    Parameters:
    ----------
    jacobian_determinants : dict
        A dictionary of jacobian determinants for different directions.
    mask : np.ndarray
        A binary mask indicating the region of interest.

    Returns:
    -------
    non_diff_area : np.ndarray
        A matrix of floating numbers that indicate the non-diffeomorphic area.
    """
    non_diff_area = np.zeros_like(jacobian_determinants['00'])
    for diff_direction in ['++', '+-', '-+', '--']:
        non_diff_area += -0.5 * np.minimum(jacobian_determinants[diff_direction], 0) * mask / 2

    return non_diff_area

def calc_non_diffeomorphic_voxels(jacobian_determinants, mask):
    """
    Calculates the number of non-diffeomorphic voxels using the given jacobian determinants and mask.

    Parameters:
    ----------
    jacobian_determinants : dict
        A dictionary of jacobian determinants for different directions.
    mask : np.ndarray
        A binary mask indicating the region of interest.

    Returns:
    -------
    non_diff_voxels : np.ndarray
        A binary mask indicating the non-diffeomorphic voxels.
    """
    non_diff_voxels = (jacobian_determinants['000'] <= 0) * mask
    return non_diff_voxels

def calc_non_diffeomorphic_pixels(jacobian_determinants, mask):
    """
    Calculates the number of non-diffeomorphic pixels using the given jacobian determinants and mask.

    Parameters:
    ----------
    jacobian_determinants : dict
        A dictionary of jacobian determinants for different directions.
    mask : np.ndarray
        A binary mask indicating the region of interest.

    Returns:
    -------
    non_diff_pixels : np.ndarray
        A binary mask indicating the non-diffeomorphic pixels.
    """
    non_diff_pixels = (jacobian_determinants['00'] <= 0) * mask
    return non_diff_pixels



import numpy as np
import argparse
import pdb
import nibabel as nib

from digital_diffeomorphism.jacobian_calculator import calc_jacobian_determinants_2d
from digital_diffeomorphism.utils import identity_grid_like
from digital_diffeomorphism.non_diffeomorphic_measures import (
    calc_non_diffeomorphic_area, calc_non_diffeomorphic_pixels)

parser = argparse.ArgumentParser(description='Calculate non-diffeomorphic area')
parser.add_argument('transformation_file', type=str, help='Path to the transformation Nifti file')
parser.add_argument('--mask_file', type=str, help='Path to the mask Nifti file')
parser.add_argument('--disp', action='store_true', help='Flag for displacement field input')
args = parser.parse_args()

# load transformation data
transformation_obj = nib.load(args.transformation_file)
transformation = transformation_obj.get_fdata()

# convert displacement field to deformation field if necessary
if args.disp:
    transformation += identity_grid_like(transformation)

jacobian_determinants = calc_jacobian_determinants_2d(transformation)

if args.mask_file:
    mask = nib.load(args.mask_file).get_fdata()
    mask = (mask > 0).astype('float32')
else:
    mask = np.ones_like(transformation[..., 0])

total_pixels = np.sum(mask)

# non-diffeomorphic area
non_diff_area = calc_non_diffeomorphic_area(jacobian_determinants, mask)
nib.save(nib.Nifti1Image(non_diff_area, transformation_obj.affine),
         args.transformation_file.replace('.nii.gz', '_non_diff_area.nii.gz'))
non_diff_area = np.sum(non_diff_area)
non_diff_area_percentage = non_diff_area / total_pixels * 100
print(f'Non-diffeomorphic Area: {non_diff_area:.2f}({non_diff_area_percentage:.2f}%)')

# non-diffeomorphic pixels
non_diff_pixels = calc_non_diffeomorphic_pixels(jacobian_determinants, mask)
nib.save(nib.Nifti1Image(non_diff_pixels, transformation_obj.affine),
         args.transformation_file.replace('.nii.gz', '_non_diff_pixels.nii.gz'))
non_diff_pixels = np.sum(non_diff_pixels)
non_diff_pixels_percentage = non_diff_pixels / total_pixels * 100
print(f'Non-diffeomorphic Pixels: {non_diff_pixels:.2f}({non_diff_pixels_percentage:.2f}%)')



import numpy as np
import argparse
import nibabel as nib
import torch

from digital_diffeomorphism.core import ndv, nda
from digital_diffeomorphism.utils import identity_grid_like, identity_grid, spatial_padding

class TestCoreNDVIdentity:
    def setup_method(self, method):
        transformation = identity_grid(shape=[3, 3, 3])

        self.out = ndv(transformation, is_disp=False)

    def test_nd_area(self):
        target = 0
        assert np.array_equal(self.out[0], target)

    def test_nd_pixels(self):
        target = 0
        assert np.array_equal(self.out[3], target)

class TestNDVDispIdentity:
    def setup_method(self, method):
        transformation = identity_grid(shape=[3, 3, 3])
        disp = transformation - identity_grid_like(transformation)

        self.out = ndv(disp, is_disp=True)

    def test_nd_area(self):
        target = 0
        assert np.array_equal(self.out[0], target)

    def test_nd_pixels(self):
        target = 0
        assert np.array_equal(self.out[3], target)

class TestCoreNDAIdentity:
    def setup_method(self, method):
        transformation = identity_grid(shape=[3, 3])

        self.out = nda(transformation, is_disp=False)

    def test_nd_area(self):
        target = 0
        assert np.array_equal(self.out[0], target)

    def test_nd_pixels(self):
        target = 0
        assert np.array_equal(self.out[3], target)

class TestNDADispIdentity:
    def setup_method(self, method):
        transformation = identity_grid(shape=[3, 3])
        disp = transformation - identity_grid_like(transformation)

        self.out = nda(disp, is_disp=True)

    def test_nd_area(self):
        target = 0
        assert np.array_equal(self.out[0], target)

    def test_nd_pixels(self):
        target = 0
        assert np.array_equal(self.out[3], target)



import numpy as np
import argparse
import nibabel as nib
import torch

from digital_diffeomorphism.jacobian_calculator_torch import (
            calc_jacdets_3d,
            calc_jacdets_2d
)
from digital_diffeomorphism.utils import identity_grid_like

def test_proposition_5_ndv():
    # Generate a random 100x100x100x3 matrix with values in the range [-10, 10]
    displacement_field = np.random.uniform(low=-10, high=10, size=(100, 100, 100, 3))
    transformation = displacement_field + identity_grid_like(displacement_field)

    # calculate the finite difference based jacobian determinants
    jacdets = calc_jacdets_3d(transformation, torch.device("cpu"))

    jacdets['intersection'] = np.ones_like(jacdets['000'])
    for direction in ['+++', '++-', '+-+', '+--', '-++', '-+-', '--+', '---']:
        jacdets['intersection'] *= (jacdets[direction] > 0)

    # if a voxel has 'intersection' positive, the central difference must be positive
    assert np.sum((jacdets['intersection'] > 0) * (jacdets['000'] <= 0)) == 0

def test_proposition_5_nda():
    # Generate a random 1000x1000x2 matrix with values in the range [-5, 5]
    displacement_field = np.random.uniform(low=-5, high=5, size=(1000, 1000, 2))
    transformation = displacement_field + identity_grid_like(displacement_field)

    # calculate the finite difference based jacobian determinants
    jacdets = calc_jacdets_2d(transformation, torch.device("cpu"))

    jacdets['intersection'] = np.ones_like(jacdets['00'])
    for direction in ['++', '+-', '-+', '--']:
        jacdets['intersection'] *= (jacdets[direction] > 0)

    # if a voxel has 'intersection' positive, the central difference must be positive
    assert np.sum((jacdets['intersection'] > 0) * (jacdets['00'] <= 0)) == 0



import pdb
import numpy as np

def identity_grid_like(grid):
    """
    Create an identity grid with the same spatial dimension as the given NumPy grid.

    Args:
        grid (np.ndarray): A 2D or 3D grid whose spatial dimension will be used to create the identity grid.

    Returns:
        identity (np.ndarray): An identity grid with the same spatial dimension as the input grid, plus an additional dimension
        for the number of dimensions of the input grid. For example, if the input grid has shape (n1, n2, n3, 3),
        then the output grid will also have shape (n1, n2, n3, 3).

    """
    shape = grid.shape[:-1]

    vectors = [np.arange(0, dim, 1) for dim in shape]

    grids = np.meshgrid(*vectors)
    grids = [np.swapaxes(x, 0, 1) for x in grids]

    identity = np.stack(grids, axis=-1).astype('float32')

    return identity

def identity_grid(shape, padding=0):
    """
    Generate an identity grid of the specified shape with optional padding.
    
    Args:
        shape (tuple): A tuple representing the dimensions of the output grid.
        padding (int, optional, default: 0): The number of paddingto add
        around the grid. 
    
    Returns:
        (numpy.ndarray): An identity grid with the specified shape and padding.
        The output array will have the same number of dimensions as the input shape, and
        the value at each position will correspond to its own index.
    """
    vectors = [np.arange(0-padding, dim+padding, 1) for dim in shape]

    grids = np.meshgrid(*vectors)
    grids = [np.swapaxes(x, 0, 1) for x in grids]

    identity = np.stack(grids, axis=-1).astype('float32')

    return identity

def spatial_padding(transformation):
    """
    Apply spatial padding to a given transformation by extending its dimensions
    and pad around its edges.
    
    Args:
        transformation (np.ndarray): A NumPy array representing the transformation.
    
    Returns:
        (numpy.ndarray): A padded transformation.
    """
    shape = transformation.shape[:-1]
    disp = transformation - identity_grid(shape=shape)

    padding = [(1, 1)] * (transformation.ndim - 1) + [(0, 0)]
    padded_disp = np.pad(disp, padding, mode='edge')

    padded_identity = identity_grid(shape=shape, padding=1)
    return padded_disp + padded_identity

    



import numpy as np
import argparse
import nibabel as nib
import torch

from digital_diffeomorphism.jacobian_calculator_torch import calc_jacdets_2d
from digital_diffeomorphism.utils import identity_grid_like, identity_grid, spatial_padding
from digital_diffeomorphism.non_diffeomorphic_measures import calc_non_diffeomorphic_area

class TestExample():
    def setup_method(self, method):
        transformation = identity_grid(shape=[3, 3])
        transformation[1, 1, :] = [0, 2]

        self.jacdets = calc_jacdets_2d(transformation, torch.device("cpu"))

    def test_central_difference(self):
        target = np.array([[1, 0.5, 1], [1.5, 1, 0.5], [1, 1.5, 1]])
        assert np.array_equal(self.jacdets['00'], target)

    def test_forward_forward_difference(self):
        target = np.array([[1, 0, 1], [2, 1, 1], [1, 1, 1]])
        assert np.array_equal(self.jacdets['++'], target)

    def test_forward_backward_difference(self):
        target = np.array([[1, 0, 1], [1, 3, 0], [1, 1, 1]])
        assert np.array_equal(self.jacdets['+-'], target)

    def test_backward_forward_difference(self):
        target = np.array([[1, 1, 1], [2, -1, 1], [1, 2, 1]])
        assert np.array_equal(self.jacdets['-+'], target)

    def test_backward_backward_difference(self):
        target = np.array([[1, 1, 1], [1, 1, 0], [1, 2, 1]])
        assert np.array_equal(self.jacdets['--'], target)

    def test_non_diffeomorphic_area(self):
        mask = np.ones(shape=[3, 3])
        non_diff_area = calc_non_diffeomorphic_area(self.jacdets, mask)
        target = np.array([[0, 0, 0], [0, 1/4, 0], [0, 0, 0]])
        assert np.array_equal(non_diff_area, target)

class TestIdentity:
    def setup_method(self, method):
        transformation = identity_grid(shape=[3, 3])

        self.jacdets = calc_jacdets_2d(transformation, torch.device("cpu"))

    def test_central_difference(self):
        target = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
        assert np.array_equal(self.jacdets['00'], target)

    def test_forward_forward_difference(self):
        target = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
        assert np.array_equal(self.jacdets['++'], target)

    def test_forward_backward_difference(self):
        target = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
        assert np.array_equal(self.jacdets['+-'], target)

    def test_backward_forward_difference(self):
        target = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
        assert np.array_equal(self.jacdets['-+'], target)

    def test_backward_backward_difference(self):
        target = np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]])
        assert np.array_equal(self.jacdets['--'], target)

    def test_non_diffeomorphic_area(self):
        mask = np.ones(shape=[3, 3])
        non_diff_area = calc_non_diffeomorphic_area(self.jacdets, mask)
        target = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        assert np.array_equal(non_diff_area, target)
